(* This program is free software; you can redistribute it and/or      *)
(* modify it under the terms of the GNU Lesser General Public License *)
(* as published by the Free Software Foundation; either version 2.1   *)
(* of the License, or (at your option) any later version.             *)
(*                                                                    *)
(* This program is distributed in the hope that it will be useful,    *)
(* but WITHOUT ANY WARRANTY; without even the implied warranty of     *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      *)
(* GNU General Public License for more details.                       *)
(*                                                                    *)
(* You should have received a copy of the GNU Lesser General Public   *)
(* License along with this program; if not, write to the Free         *)
(* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA *)
(* 02110-1301 USA                                                     *)


Require Import Termes.
Require Import Conv.
Require Import Types.
Require Import Class.
Require Import Can.
Require Import Int_term.
Require Import Int_typ.
Require Import Int_stab.

Load "ImpVar".

Inductive trm_in_int : env -> intP -> intt -> Prop :=
  | int_nil : forall itt : intt, trm_in_int nil (TNl _) itt
  | int_cs :
      forall e (ip : intP) (itt : intt),
      trm_in_int e ip itt ->
      forall (y : Int_K) t T,
      int_typ T ip PROP t ->
      trm_in_int (T :: e) (TCs _ y ip) (shift_intt itt t).

  Hint Resolve int_nil int_cs: coc.


  Record int_adapt e (ip : intP) (itt : intt) : Prop := 
    {adapt_trm_in_int : trm_in_int e ip itt;
     int_can_adapt : can_adapt ip;
     adapt_class_equal : cls_of_int ip = class_env e}.

(* BEAUTY: this lemma goes to types.v *)

  Lemma typ_prop_set {e A B} : typ e A B -> typ e (Srt prop) (Srt set).
intros. constructor.
now apply typ_wf with A B.
Qed.


  Lemma sn_substitute_intt B it : sn (int_term B (shift_intt it (Ref 0)) 0) -> sn (int_term B it 1).
intros.
apply sn_subst with (Ref 0).
unfold subst. now rewrite int_term_subst.
Qed.

  Lemma extend_int A e ip it : typ e A (Srt prop) -> int_adapt e ip it -> int_adapt (A :: e) (def_cons A ip) (shift_intt it (Ref 0)).

intros typA [adapt_trm_in_int int_can_adapt adapt_class_equal]. 
unfold def_cons, int_cons ; simpl.
apply Build_int_adapt.
  apply int_cs ; superauto.
  apply (var_in_cand 0 (int_typ A ip PROP)) ; superauto.
  
  red. constructor ; superauto.
  unfold ext_ik. destruct (cl_term A (cls_of_int ip)) ; superauto.
  
  (* BEAUTY : make a lemma for this crap... it's a fucking leitmotiv *)
  unfold ext_ik. rewrite adapt_class_equal.
  unfold cls_of_int. simpl.
  assert (H' := class_sound e A (Srt prop) typA (Srt set) (typ_prop_set typA)).
  simpl in H' ; inversion_clear H'.
  rewrite <- adapt_class_equal ; superauto.
Qed.

  Lemma sn_extended_int A B e ip it : typ e A (Srt prop) -> int_adapt e ip it -> (forall (ip : intP) (it : intt),
         int_adapt (A :: e) ip it -> sn (int_term B it 0)) -> sn (int_term B it 1).

intros. apply sn_substitute_intt.
apply (H1 _ _ (extend_int A e ip it H H0)).
Qed.

Hint Resolve sn_substitute_intt extend_int sn_extended_int : coc.
 
  Lemma int_sound :
   forall e t T,
   typ e t T ->
   forall (ip : intP) (it : intt),
   int_adapt e ip it -> int_typ T ip PROP (int_term t it 0).
induction 1; intros.

(* prop *)
red in |- *; apply Acc_intro; intros.
inversion_clear H1.

(* Ref *)
simpl.
elim (le_gt_dec 0 v); [ intro Hle | intro Hgt ].
rewrite lift0.
elim minus_n_O with v.
elim H0; intros.
rewrite H2.
generalize ip it H1.
clear H2 Hle H1 it ip H0 t.
elim H3.
intros l ip it (in_interp);
  (* Do not intro other fields *)
 inversion_clear in_interp; simpl in |- *; intros ip_can_adapted same_classes.
apply eq_cand_incl with (int_typ x ip0 PROP);
 auto with coc core arith datatypes.
unfold lift in |- *.
apply lift_int_typ with y; auto with coc core arith datatypes.

intros l n y H1 H2 ip it (in_interp); inversion_clear in_interp;
 simpl in |- *; intros ip_can_adapted same_classes.
simpl in |- *.
rewrite simpl_lift.
apply eq_cand_incl with (int_typ (lift (S n) x) ip0 PROP).
unfold lift at 2 in |- *.
apply lift_int_typ with y0; auto with coc core arith datatypes.

apply H2.
apply Build_int_adapt; auto with coc core arith datatypes.
inversion_clear ip_can_adapted; auto with coc core arith datatypes.

injection same_classes; auto with coc core arith datatypes.

inversion_clear Hgt.

(* Abs *)

elim H2; intros in_interp ip_can_adapted same_classes.
simpl.
apply Abs_sound; intros; auto with coc core arith datatypes.
simpl in |- *; intros.
change
  (is_can PROP
     (int_typ U (int_cons T ip (cv_skel (cl_term T (cls_of_int ip))) X)
        PROP)) in |- *.
apply int_typ_cr.
unfold int_cons, ext_ik in |- *.
generalize X H3 H4.
elim (cl_term T (cls_of_int ip)); auto with coc core arith datatypes.

unfold subst in |- *.
rewrite int_term_subst; auto with coc core arith datatypes.
apply IHtyp3.
unfold int_cons, ext_ik in |- *.
apply Build_int_adapt; auto with coc core arith datatypes.
generalize C H4 H5.
elim (cl_term T (cls_of_int ip)); auto with coc core arith datatypes.

simpl in |- *.
pattern (cls_of_int ip) at 1 in |- *.
rewrite same_classes.
unfold cls_of_int in |- *.
pattern (cl_term T (class_env e)) in |- *.
apply class_typ_ord with s1; elim same_classes; simpl in |- *;
 auto with coc core arith datatypes.
rewrite same_classes.
elim skel_sound with e T (Srt s1); simpl in |- *;
 auto with coc core arith datatypes.
elim same_classes; auto with coc core arith datatypes.

apply IHtyp1 with ip; auto with coc core arith datatypes.

(* App *)

simpl. simpl in IHtyp2.
elim H1; intros in_interp ip_can_adapted same_classes.
elim type_case with e u (Prod V Ur); intros;
 auto with coc core arith datatypes.
inversion_clear H2.
apply inv_typ_prod with e V Ur (Srt x); auto with coc core arith datatypes;
 intros.
apply
 eq_cand_incl
  with
    (int_typ Ur
       (int_cons V ip (cv_skel (cl_term V (cls_of_int ip))) (int_typ v ip _))
       PROP).
replace PROP with
 (skel_int Ur
    (int_cons V ip (cv_skel (cl_term V (cls_of_int ip)))
       (int_typ v ip (cv_skel (cl_term V (cls_of_int ip)))))).
unfold subst, int_cons in |- *.
apply
 subst_int_typ
  with
    ip
    (ext_ik V ip (cv_skel (cl_term V (cls_of_int ip)))
       (int_typ v ip (cv_skel (cl_term V (cls_of_int ip)))))
    (V :: e)
    (Srt s2); auto with coc core arith datatypes.
unfold ext_ik in |- *.
rewrite same_classes.
cut (cl_term v (cls_of_int ip) = cl_term v (class_env e)).
elim class_sound with e v V (Srt s1); intros;
 auto with coc core arith datatypes.

elim same_classes; auto with coc core arith datatypes.

simpl in |- *.
unfold ext_ik in |- *.
rewrite same_classes.
unfold cls_of_int in |- *.
apply class_typ_ord with s1; elim same_classes; simpl in |- *;
 auto with coc core arith datatypes.
rewrite same_classes.
elim skel_sound with e V (Srt s1); simpl in |- *;
 auto with coc core arith datatypes.
elim same_classes; auto with coc core arith datatypes.

unfold ext_ik in |- *.
red in |- *; red in |- *; auto with coc core arith datatypes.
apply Tfa2_cs; auto with coc core arith datatypes.
elim (cl_term V (cls_of_int ip)); auto with coc core arith datatypes.

change (int_inv ip) in |- *.
apply adapt_int_inv; auto with coc core arith datatypes.

replace
 (cls_of_int
    (TCs Int_K
       (ext_ik V ip (cv_skel (cl_term V (cls_of_int ip)))
          (int_typ v ip (cv_skel (cl_term V (cls_of_int ip))))) ip)) with
 (class_env (V :: e)).
apply class_typ_ord with s2; auto with coc core arith datatypes.
discriminate.

discriminate.

simpl in |- *.
unfold cls_of_int at 1, ext_ik in |- *.
rewrite same_classes.
pattern (cl_term V (class_env e)) in |- *.
apply class_typ_ord with s1; elim same_classes; simpl in |- *;
 auto with coc core arith datatypes.
rewrite same_classes.
elim skel_sound with e V (Srt s1); simpl in |- *;
 auto with coc core arith datatypes.
elim same_classes; auto with coc core arith datatypes.

unfold int_cons, skel_int in |- *.
replace
 (cls_of_int
    (TCs Int_K
       (ext_ik V ip _ (int_typ v ip (cv_skel (cl_term V (cls_of_int ip)))))
       ip)) with (class_env (V :: e)).
elim skel_sound with (V :: e) Ur (Srt s2); simpl in |- *;
 auto with coc core arith datatypes.

simpl in |- *.
unfold ext_ik in |- *.
rewrite same_classes.
unfold cls_of_int in |- *.
elim class_sound with e v V (Srt s1); auto with coc core arith datatypes.
simpl in |- *.
elim same_classes; auto with coc core arith datatypes.

simpl in |- *.
elim same_classes; auto with coc core arith datatypes.

unfold Pi in IHtyp2.
apply IHtyp2; auto with coc core arith datatypes.

discriminate.

(* Prod *)

simpl.
apply sn_prod.
apply IHtyp1 with ip; auto with coc core arith datatypes.

apply sn_subst with (Ref 0).
unfold subst in |- *.
rewrite int_term_subst.
elim H1; intros in_interp ip_can_adapted same_classes.
apply IHtyp2 with (def_cons T ip).
unfold def_cons, int_cons in |- *.
apply Build_int_adapt.
apply int_cs; auto with coc core arith datatypes.
apply (var_in_cand 0 (int_typ T ip PROP));
 auto with coc core arith datatypes.

red in |- *.
apply Tfa_cs; auto with coc core arith datatypes.
unfold ext_ik in |- *.
elim (cl_term T (cls_of_int ip)); auto with coc core arith datatypes.

unfold ext_ik in |- *.
rewrite same_classes.
unfold cls_of_int in |- *.
simpl in |- *.
pattern (cl_term T (class_env e)) in |- *.
apply class_typ_ord with s1; simpl in |- *; elim same_classes;
 auto with coc core arith datatypes.
rewrite same_classes.
elim skel_sound with e T (Srt s1); simpl in |- *;
 auto with coc core arith datatypes.
elim same_classes; auto with coc core arith datatypes.

cut (typ e U (Srt s)); auto with coc core arith datatypes.
intros.
apply eq_cand_incl with (int_typ U ip PROP);
 auto with coc core arith datatypes.
replace PROP with (skel_int U ip).
elim H2; intros in_interp ip_can_adapted same_classes.
apply conv_int_typ with e (Srt s); auto with coc core arith datatypes.
apply class_typ_ord with s; auto with coc core arith datatypes.
discriminate.

discriminate.

unfold skel_int in |- *.
elim H2; intros in_interp ip_can_adapted same_classes.
rewrite same_classes.
elim skel_sound with e U (Srt s); simpl in |- *;
 auto with coc core arith datatypes.

elim type_case with e t U; intros; auto with coc core arith datatypes.
inversion_clear H3.
elim conv_sort with x s; auto with coc core arith datatypes.
apply typ_conv_conv with e U V; auto with coc core arith datatypes.

rewrite H3 in H0.
elim inv_typ_conv_set with e V (Srt s); auto with coc core arith datatypes.

(* Sigma *)

simpl.
apply sn_sigma.
  now apply IHtyp1 with ip.
  
  apply sn_extended_int with A e ip ; superauto.

(* Pair *)    

elim H3 ; intros in_interp ip_can_adapted same_classes.
simpl.

(* typ soundness *)
pose (cl_snd := fun {e A} H => class_sound e A (Srt prop) H (Srt set) (typ_prop_set H)).
assert (clA := cl_snd _ _ H).
assert (clB := cl_snd _ _ H0).

(* term soundness *)
pose (tm_snd := fun {e A B} H1 H2 => class_sound e A B H1 (Srt prop) H2).
assert (clM := tm_snd _ _ _ H1 H).

(* we know the class of prop *)
simpl in clA, clB.
inversion clA ; inversion clB ; subst.
(* now we know the class of A *)
rewrite <- H5 in clM. inversion clM.

remember (int_cons A ip _ (int_typ A ip (cv_skel (cl_term A (class_env e))))) as ext_int.
assert (can_adapt ext_int).
  rewrite Heqext_int ; constructor ; superauto.
  unfold ext_ik. rewrite same_classes, <- H5 ; superauto.
assert (ext_same_classes : cls_of_int ext_int = class_env (A :: e)).
  rewrite Heqext_int. rewrite int_cons_equal with _ _ _ prop _ ; superauto.

assert (canA : is_can PROP (int_typ A ip PROP)) by superauto.
assert (canB : is_can PROP (int_typ B ext_int PROP)) by superauto.


  (* took me a while I could "specialize" like this *)	
assert (subst_int := fun H1 H2 => subst_int_typ M ip iT H1 (A :: e) B (Srt prop) H2 0 ext_int ip).

(* ext_int is no longer useful... die. *)
subst.

apply cartesian_intro ; superauto.
  pattern PROP at 1 2.
  replace PROP with (cv_skel (cl_term A (class_env e))) ; superauto.
  rewrite <- H5 ; superauto.
  
  apply IHtyp1  with ip ; superauto.
  
  apply sn_extended_int with A e ip ; superauto.
  
  unfold skel_int in subst_int.
  rewrite ! ext_same_classes in subst_int.
  simpl in subst_int.
  rewrite <- H7, <- H5 in subst_int.
  apply subst_int ; superauto.
    constructor. rewrite same_classes ; superauto. (* faster than now oO *)
    
    unfold int_cons, ext_ik. rewrite same_classes, <- H5 ; superauto.
    
    unfold int_cons, ext_ik. rewrite same_classes, <- H5 ; superauto.
    
    discriminate.

(* Proj1 *)
simpl in *.
elim (IHtyp ip it H0) ; superauto.

(* Proj2 *)

elim H0; intros in_interp ip_can_adapted same_classes.
simpl in *.

apply inv_typ_sigma with e M A B ; superauto ; intros.

(* typ soundness *)
pose (cl_snd := fun {e A} H => class_sound e A (Srt prop) H (Srt set) (typ_prop_set H)).
assert (clA := cl_snd _ _ H1).
assert (clB := cl_snd _ _ H2).

(* we know the class of prop *)
simpl in clA, clB.
inversion clA ; inversion clB ; subst.


remember (int_cons A ip _ (int_typ A ip PROP)) as ext_int.
assert (can_adapt ext_int).
  rewrite Heqext_int ; constructor ; superauto.
  unfold ext_ik. rewrite same_classes, <- H5 ; superauto.
assert (ext_same_classes : cls_of_int ext_int = class_env (A :: e)).
  rewrite Heqext_int. 
  replace PROP with (cv_skel (cl_term A (class_env e))).
  rewrite int_cons_equal with _ _ _ prop _ ; superauto.
  rewrite <- H5 ; superauto.

assert (canA : is_can PROP (int_typ A ip PROP)) by superauto.
assert (canB : is_can PROP (int_typ B ext_int PROP)) by superauto.


(* same argument as above applies here, just the other direction. *)

assert (subst_int := fun H1 H2 => subst_int_typ (Proj1 M) ip iT H1 (A :: e) B (Srt prop) H2 0 ext_int ip).

(* ext_int has served us enough... time to die. *)
subst.

unfold skel_int in subst_int.
rewrite ! ext_same_classes in subst_int.
simpl in subst_int.
rewrite <- H7 in subst_int.
apply subst_int ; superauto.
  unfold int_cons, ext_ik. rewrite same_classes, <- H5.
  now constructor.
  
  discriminate.
  
  apply IHtyp ; superauto.
Qed.



  Fixpoint def_intp e : intP :=
    match e with
    | nil => TNl _
    | t :: f => def_cons t (def_intp f)
    end.



  Fixpoint def_intt e : nat -> intt :=
    fun k =>
    match e with
    | nil => fun p => Ref (k + p)
    | _ :: f => shift_intt (def_intt f (S k)) (Ref k)
    end.


  Lemma def_intp_can : forall e, can_adapt (def_intp e).
simple induction e; simpl in |- *; auto with coc core arith datatypes; intros.
unfold def_cons, int_cons, ext_ik in |- *.
elim (cl_term a (cls_of_int (def_intp l)));
 auto with coc core arith datatypes.
Qed.


  Lemma def_adapt :
   forall e, wf e -> forall k, int_adapt e (def_intp e) (def_intt e k).
simple induction e; simpl in |- *; intros.
apply Build_int_adapt; auto with coc core arith datatypes.

inversion_clear H0.
cut (wf l); intros.
elim H with (S k); trivial; intros in_interp ip_can_adapted same_classes.
unfold def_cons, int_cons in |- *.
apply Build_int_adapt; auto with coc core arith datatypes.
apply int_cs; auto with coc core arith datatypes.
apply (var_in_cand k (int_typ a (def_intp l) PROP));
 auto with coc core arith datatypes.

unfold ext_ik in |- *.
rewrite same_classes.
pattern (cl_term a (class_env l)) in |- *.
apply class_typ_ord with s; auto with coc core arith datatypes.

simpl in |- *.
unfold ext_ik in |- *.
rewrite same_classes.
pattern (cl_term a (class_env l)) in |- *.
apply class_typ_ord with s; unfold cls_of_int in |- *; elim same_classes;
 auto with coc core arith datatypes.
simpl in |- *.
rewrite same_classes.
elim skel_sound with l a (Srt s); auto with coc core arith datatypes.
simpl in |- *; auto with coc core arith datatypes.
elim same_classes; auto with coc core arith datatypes.

apply typ_wf with a (Srt s); auto with coc core arith datatypes.
Qed.

  Hint Resolve def_intp_can def_adapt: coc.


  Lemma def_intt_id : forall n e k, def_intt e k n = Ref (k + n).
simple induction n; simple destruct e; simpl in |- *;
 auto with coc core arith datatypes; intros.
replace (k + 0) with k; auto with coc core arith datatypes.

rewrite H.
replace (k + S n0) with (S (k + n0)); auto with coc core arith datatypes.
Qed.


  Lemma id_int_term : forall e t k, int_term t (def_intt e 0) k = t.
simple induction t; simpl in |- *; intros; auto with coc core arith datatypes.
elim (le_gt_dec k n); intros; auto with coc core arith datatypes.
rewrite def_intt_id.
simpl in |- *; unfold lift in |- *.
rewrite lift_ref_ge; auto with coc core arith datatypes.


rewrite H; rewrite H0; auto with coc core arith datatypes.

rewrite H; rewrite H0; auto with coc core arith datatypes.

rewrite H; rewrite H0; auto with coc core arith datatypes.

rewrite H, H0 ; superauto.

rewrite H, H0, H1, H2 ; superauto.

rewrite H ; superauto.

rewrite H ; superauto.
Qed.




  Theorem str_norm : forall e t T, typ e t T -> sn t.
intros.
cut (is_can PROP (int_typ T (def_intp e) PROP));
 auto with coc core arith datatypes.
simpl in |- *; intros.
cut (int_typ T (def_intp e) PROP t).
elim H0; auto with coc core arith datatypes.

elim id_int_term with e t 0.
apply int_sound with e; auto with coc core arith datatypes.
apply def_adapt.
apply typ_wf with t T; auto with coc core arith datatypes.
Qed.



  Lemma type_sn : forall e t T, typ e t T -> sn T.
intros.
elim type_case with e t T; intros; auto with coc core arith datatypes.
elim H0; intros.
apply str_norm with e (Srt x); auto with coc core arith datatypes.

rewrite H0.
red in |- *; apply Acc_intro; intros.
inversion_clear H1.
Qed.
